package org.gvsig.mobile.core.concepts;

import java.io.FileInputStream;
import java.io.IOException;
import java.net.MalformedURLException;

import org.gvsig.mobile.core.exceptions.GvSigPropertyNotAttachedException;
import org.gvsig.mobile.core.extensions.interfaces.IExtension;
import org.gvsig.mobile.core.extensions.interfaces.IFunctionality;
import org.gvsig.mobile.core.extensions.interfaces.ITool;
import org.gvsig.mobile.core.extensions.interfaces.IToolbar;
import org.gvsig.mobile.ui.awt.ImageUtils;
import org.gvsig.mobile.ui.lw.IconResourceReader;
import org.kxml2.io.KXmlParser;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.xmlpull.v1.XmlPullParserException;

/**
 * Any bundle activator should extend this class. It reads the configuration
 * file from the bundle folder to register any extensions, tools and functionalities 
 * into the OSGi framework.
 * 
 * @author aromeu
 * 
 */
public class BaseActivator implements BundleActivator {

	private final static String BUNDLE_CONFIG = "bundle-config";

	private final static String EXTENSIONS = "extensions";
	private final static String TOOLBARS = "tool-bars";
	private final static String TOOLS = "tools";
	private final static String FUNCTIONALITIES = "functionalities";

	private final static String EXTENSION = "extension";
	private final static String TOOLBAR = "tool-bar";
	private final static String TOOL = "tool";
	private final static String FUNCTIONALITY = "functionality";

	private final static String CLASS_NAME = "class-name";
	private final static String ID = "ID";
	private final static String NAME = "name";
	private final static String VERSION = "version";
	private final static String PROVIDER = "provider";
	private final static String PUBLICATION_DATE = "publicationDate";
	private final static String DESCRIPTION = "description";
	private final static String COMMENTS = "comments";

	private final static String KEY = "key";
	private final static String ICON = "icon";
	private final static String TEXT = "text";
	private final static String SHORT_DESC = "shortDesc";
	private final static String EXTENSION_ID = "extension-ID";
	private final static String FUNC_ID = "func-ID";
	private final static String FUNC_NAME = "func-name";
	private final static String FUNC_VERSION = "func-version";
	private final static String FUNC_PROVIDER = "func-provider";
	private final static String FUNC_EXT_ID = "func-ext-ID";
	private final static String FUNC_INTERFACE_ID = "func-interface-ID";
	private final static String FUNC_INTERFACE_NAME = "func-interface-name";
	private final static String FUNC_INTERFACE_VERSION = "func-interface-version";

	private final static String POSITION = "position";
	private final static String TOOLBAR_KEY = "toolbar-key";
	private final static String TOOL_TYPE = "tool-type";
	
	private final static String INTERFACE_ID = "interface-ID";
	private final static String INTERFACE_NAME = "interface-name";
	private final static String INTERFACE_VERSION = "interface-version";
	
	private final static String EVENT_SUBSCRIPTIONS = "event-subscriptions";
	private final static String MOUSE_EVENT_SUBSCRIPTION = "mouse-event-subscription";
	
	private final static String CONFIG_FILE = "config.xml";

	private String location;
	protected Bundle bundle;
	protected BundleContext context;	
	
	/**
	 * This method must be called by any activator when is started by the
	 * framework
	 * @param context The BundleContext
	 * @param bundle The current bundle
	 * @throws Exception
	 */
	public void start(BundleContext context, Bundle bundle) throws Exception {
		this.context = context;
		this.bundle = bundle;
		configureLocale();
		_start(context);
	}
	
	private void configureLocale() {
		try {			
			String name = bundle.getSymbolicName();
			org.gvsig.i18n.Messages.addResourceFamily("text", IconResourceReader.getLanguageFolder(name, "languages"));		
		} catch (MalformedURLException e) {
			System.err.println("Language files not found in resources directory.");
		}
	}

	private void _start(BundleContext context) throws Exception {
		
		location = context.getBundle().getLocation();
		
		
		location = ImageUtils.parseBundleLocation(location);
		
		//Get config.xml
		FileInputStream f = new FileInputStream(
		location.toString() + CONFIG_FILE);
//		location.substring(5, location.length() - 4) + File.separator + "config.xml");

		//Initialize the parser
		KXmlParser kxmlParser = new KXmlParser();
		kxmlParser.setInput(f, "ISO-8859-1");

		int tag;
		IExtension extension = null;

		try {
			kxmlParser.nextTag();
			if (kxmlParser.getEventType() != KXmlParser.END_DOCUMENT) {
				kxmlParser.require(KXmlParser.START_TAG, null, BUNDLE_CONFIG);
				tag = kxmlParser.nextTag();
				while (tag != KXmlParser.END_DOCUMENT) {
					switch (tag) {

					case KXmlParser.START_TAG:
						if (kxmlParser.getName().compareTo(EXTENSIONS) == 0) {
							// parseServiceTag(kxmlParser);
							extension = parseExtensions(kxmlParser);
						}
						break;
					case KXmlParser.END_TAG:
						break;
					case KXmlParser.TEXT:
						break;
					}
					tag = kxmlParser.next();
				}
				kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
				f.close();
			}
		} catch (XmlPullParserException parser_ex) {
			// TODO log this
			System.err.println(parser_ex.getMessage());
		} catch (IOException ioe) {
			// TODO log this
			System.err.println(ioe.getMessage());
		} finally {
			if (extension != null) {				
				extension.registerAll();
			}
		}		
	}

	/**
	 * Does nothing
	 */
	public void stop(BundleContext context) throws Exception {
		// TODO Auto-generated method stub

	}

	private IExtension parseExtensions(KXmlParser parser) throws IOException,
			XmlPullParserException, IllegalAccessException,
			ClassNotFoundException, InstantiationException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, EXTENSIONS);
		currentTag = parser.next();
		IExtension ext = null;

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(EXTENSION) == 0) {
					ext = parseExtensionAttributes(parser);
				} else if (parser.getName().compareTo(TOOLBARS) == 0) {
					parseToolBars(parser, ext);					
				} else if (parser.getName().compareTo(TOOLS) == 0) {
					parseTools(parser, ext);
				} else if (parser.getName().compareTo(FUNCTIONALITIES) == 0) {
					parseFunctionalities(parser, ext);
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(EXTENSIONS) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		return ext;
	}

	private IExtension parseExtensionAttributes(KXmlParser parser)
			throws IOException, XmlPullParserException, InstantiationException,
			IllegalAccessException, ClassNotFoundException {

		IExtension ext = null;

		String value = parser.getAttributeValue("", CLASS_NAME);
		if (value != null && value.trim().compareTo("") != 0)
			ext = (IExtension)bundle.loadClass(value).newInstance();
		else
			ext = (IExtension)bundle.loadClass("org.gvsig.mobile.core.concepts.Extension").newInstance();
		
		ext.initializeCollections();

		value = parser.getAttributeValue("", ID);
		if (value != null)
			ext.setID(value);

		value = parser.getAttributeValue("", NAME);
		if (value != null)
			ext.setName(value);

		value = parser.getAttributeValue("", VERSION);
		if (value != null)
			ext.setVersion(value);

		value = parser.getAttributeValue("", PROVIDER);
		if (value != null)
			ext.setProvider(value);

		value = parser.getAttributeValue("", PUBLICATION_DATE);
		if (value != null)
			ext.setPublicationDate(value);

		value = parser.getAttributeValue("", DESCRIPTION);
		if (value != null)
			ext.setPublicationDate(value);

		value = parser.getAttributeValue("", COMMENTS);
		if (value != null)
			ext.setComments(value);

		ext.setBundleContext(context);
		return ext;
	}

	private IToolbar parseToolBars(KXmlParser parser, IExtension extension)
			throws IOException, XmlPullParserException, InstantiationException,
			IllegalAccessException, ClassNotFoundException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, TOOLBARS);
		currentTag = parser.next();
		IToolbar tb = null;

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(TOOLBAR) == 0) {
					tb = parseToolbarAttributes(parser);
					if (tb != null && extension != null) {
						try {
							IFunctionality fun = new DefaultToolBarFunctionality();
							fun.setID(tb.getKey());
							fun.setBundleContext(context);
							fun.setVersion("1.0.0");
							fun.setInterfaceVersion("1.0.0");
							extension.addFunctionality(fun);
							tb.setFunctionalityID(tb.getKey());
						} catch (GvSigPropertyNotAttachedException e) {
						}
						extension.addTool(tb);
					}
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(TOOLBARS) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		return tb;
	}

	private IToolbar parseToolbarAttributes(KXmlParser parser)
			throws IOException, XmlPullParserException, InstantiationException,
			IllegalAccessException, ClassNotFoundException {

		IToolbar ext = null;

		String value = parser.getAttributeValue("", CLASS_NAME);
		if (value != null && value.trim().compareTo("") != 0)
			ext = (IToolbar) bundle.loadClass(value).newInstance();
		else
			ext = (IToolbar) bundle.loadClass("org.gvsig.mobile.core.concepts.DefaultToolBar").newInstance();
		
		value = parser.getAttributeValue("", KEY);
		if (value != null)
			ext.setKey(value);

		value = parser.getAttributeValue("", ICON);
		if (value != null)
			ext.setIcon(value);

		value = parser.getAttributeValue("", TEXT);
		if (value != null)
			ext.setText(value);

		value = parser.getAttributeValue("", SHORT_DESC);
		if (value != null)
			ext.setShortDescription(value);

		value = parser.getAttributeValue("", NAME);
		if (value != null)
			ext.setName(value);

		value = parser.getAttributeValue("", EXTENSION_ID);
		if (value != null)
			ext.setExtensionID(value);

		value = parser.getAttributeValue("", FUNC_ID);
		if (value != null)
			ext.setFunctionalityID(value);

		value = parser.getAttributeValue("", FUNC_NAME);
		if (value != null)
			ext.setFunctionalityName(value);

		value = parser.getAttributeValue("", FUNC_VERSION);
		if (value != null)
			ext.setFunctionalityVersion(value);

		value = parser.getAttributeValue("", FUNC_PROVIDER);
		if (value != null)
			ext.setFunctionalityProvider(value);

		value = parser.getAttributeValue("", FUNC_EXT_ID);
		if (value != null)
			ext.setFunctionalityExtensionID(value);

		value = parser.getAttributeValue("", FUNC_INTERFACE_ID);
		if (value != null)
			ext.setFunctionalityInterfaceID(value);

		value = parser.getAttributeValue("", FUNC_INTERFACE_NAME);
		if (value != null)
			ext.setFunctionalityInterfaceName(value);

		value = parser.getAttributeValue("", FUNC_INTERFACE_VERSION);
		if (value != null)
			ext.setFunctionalityInterfaceVersion(value);

		ext.setBundleContext(context);
		return ext;
	}

	private ITool parseTools(KXmlParser parser, IExtension extension)
			throws IOException, XmlPullParserException, InstantiationException,
			IllegalAccessException, ClassNotFoundException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, TOOLS);
		currentTag = parser.next();
		ITool ext = null;

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(TOOL) == 0) {
					ext = parseToolAttributes(parser);
					if (ext != null && extension != null) {
						extension.addTool(ext);
					}
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(TOOLS) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		return ext;
	}

	private ITool parseToolAttributes(KXmlParser parser) throws IOException,
			XmlPullParserException, InstantiationException,
			IllegalAccessException, ClassNotFoundException {

		ITool ext = null;

		String value = parser.getAttributeValue("", CLASS_NAME);
		if (value != null && value.trim().compareTo("") != 0)
			ext = (ITool) bundle.loadClass(value).newInstance();
		else
			ext = (ITool) bundle.loadClass("org.gvsig.mobile.core.concepts.DefaultTool").newInstance();			

		value = parser.getAttributeValue("", POSITION);
		if (value != null && value.trim().compareTo("") != 0)
			ext.setPosition(Integer.valueOf(value).intValue());

		value = parser.getAttributeValue("", KEY);
		if (value != null)
			ext.setKey(value);

		value = parser.getAttributeValue("", ICON);
		if (value != null)
			ext.setIcon(value);

		value = parser.getAttributeValue("", TEXT);
		if (value != null)
			ext.setText(value);

		value = parser.getAttributeValue("", SHORT_DESC);
		if (value != null)
			ext.setShortDescription(value);

		value = parser.getAttributeValue("", TOOLBAR_KEY);
		if (value != null)
			ext.setToolbarKey(value);

		value = parser.getAttributeValue("", TOOL_TYPE);
		if (value != null && value.trim().compareTo("") != 0)
			ext.setToolType(Integer.valueOf(value).intValue());

		value = parser.getAttributeValue("", NAME);
		if (value != null)
			ext.setName(value);

		value = parser.getAttributeValue("", EXTENSION_ID);
		if (value != null)
			ext.setExtensionID(value);

		value = parser.getAttributeValue("", FUNC_ID);
		if (value != null)
			ext.setFunctionalityID(value);

		value = parser.getAttributeValue("", FUNC_NAME);
		if (value != null)
			ext.setFunctionalityName(value);

		value = parser.getAttributeValue("", FUNC_VERSION);
		if (value != null)
			ext.setFunctionalityVersion(value);

		value = parser.getAttributeValue("", FUNC_PROVIDER);
		if (value != null)
			ext.setFunctionalityProvider(value);

		value = parser.getAttributeValue("", FUNC_EXT_ID);
		if (value != null)
			ext.setFunctionalityExtensionID(value);

		value = parser.getAttributeValue("", FUNC_INTERFACE_ID);
		if (value != null)
			ext.setFunctionalityInterfaceID(value);

		value = parser.getAttributeValue("", FUNC_INTERFACE_NAME);
		if (value != null)
			ext.setFunctionalityInterfaceName(value);

		value = parser.getAttributeValue("", FUNC_INTERFACE_VERSION);
		if (value != null)
			ext.setFunctionalityInterfaceVersion(value);

		ext.setBundleContext(context);
		return ext;
	}

	private IFunctionality parseFunctionalities(KXmlParser parser,
			IExtension extension) throws IOException, XmlPullParserException, InstantiationException, IllegalAccessException, ClassNotFoundException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, FUNCTIONALITIES);
		currentTag = parser.next();
		IFunctionality ext = null;

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(FUNCTIONALITY) == 0) {
					ext = parseFunctionalityAttributes(parser);					
					if (ext != null && extension != null) {
						extension.addFunctionality(ext);
					}
					//TODO: PARSE EVENT-SUBSCRIPTIONS TAG
				} else if (parser.getName().compareTo(EVENT_SUBSCRIPTIONS) == 0) {
					parseEventSubscriptions(parser, ext);
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(FUNCTIONALITIES) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		return ext;
	}

	private IFunctionality parseFunctionalityAttributes(KXmlParser parser) throws IOException,
			XmlPullParserException, InstantiationException,
			IllegalAccessException, ClassNotFoundException {

		IFunctionality ext = null;

		String value = parser.getAttributeValue("", CLASS_NAME);
		if (value != null)
			ext = (IFunctionality) bundle.loadClass(value).newInstance();
		else
			return ext;		

		value = parser.getAttributeValue("", ID);
		if (value != null)
			ext.setID(value);

		value = parser.getAttributeValue("", NAME);
		if (value != null)
			ext.setName(value);
		
		value = parser.getAttributeValue("", VERSION);
		if (value != null)
			ext.setVersion(value);
		
		value = parser.getAttributeValue("", PROVIDER);
		if (value != null)
			ext.setProvider(value);

		value = parser.getAttributeValue("", EXTENSION_ID);
		if (value != null)
			ext.setExtensionID(value);

		value = parser.getAttributeValue("", INTERFACE_ID);
		if (value != null)
			ext.setInterfaceID(value);
		
		value = parser.getAttributeValue("", INTERFACE_NAME);
		if (value != null)
			ext.setInterfaceName(value);
		
		value = parser.getAttributeValue("", INTERFACE_VERSION);
		if (value != null)
			ext.setInterfaceVersion(value);
		
		value = parser.getAttributeValue("", DESCRIPTION);
		if (value != null)
			ext.setDescription(value);
		
		ext.setBundleContext(context);
		return ext;
	}
	
	private IFunctionality parseEventSubscriptions(KXmlParser parser,
			IFunctionality functionality) throws IOException, XmlPullParserException, InstantiationException, IllegalAccessException, ClassNotFoundException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, EVENT_SUBSCRIPTIONS);
		currentTag = parser.next();
		

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(MOUSE_EVENT_SUBSCRIPTION) == 0) {
					String value = parser.getAttributeValue("", "type");
					if (value != null && value.trim().compareTo("") != 0) {
						functionality.addMouseEventSubscription(Integer.valueOf(value).intValue());
					}
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(EVENT_SUBSCRIPTIONS) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		return functionality;
	}

	/**
	 * Does nothing
	 */
	public void start(BundleContext context) throws Exception {
		// TODO Auto-generated method stub
		
	}
}
