/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2004 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *  Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *      +34 963862235
 *   gvsig@gva.es
 *      www.gvsig.gva.es
 *
 *    or
 *
 *   IVER T.I. S.A
 *   Salamanca 50
 *   46005 Valencia
 *   Spain
 *
 *   +34 963163400
 *   dac@iver.es
 */
package org.gvsig.mobile.extension.wms.fmap.layers;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.gvsig.compat.CompatLocator;
import org.gvsig.compat.print.PrintAttributes;
import org.gvsig.fmap.crs.CRSFactory;
import org.gvsig.fmap.dal.exception.ReadException;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.Geometry.SUBTYPES;
import org.gvsig.fmap.geom.exception.CreateEnvelopeException;
import org.gvsig.fmap.geom.primitive.Envelope;
import org.gvsig.fmap.mapcontext.ViewPort;
import org.gvsig.fmap.mapcontext.exceptions.ConnectionErrorLayerException;
import org.gvsig.fmap.mapcontext.exceptions.LoadLayerException;
import org.gvsig.fmap.mapcontext.exceptions.URLLayerException;
import org.gvsig.fmap.mapcontext.exceptions.UnsupportedVersionLayerException;
import org.gvsig.fmap.mapcontext.layers.FLayer;
import org.gvsig.fmap.mapcontext.layers.FLyrDefault;
import org.gvsig.fmap.mapcontext.layers.operations.IHasImageLegend;
import org.gvsig.fmap.mapcontext.rendering.legend.ILegend;
import org.gvsig.mobile.core.auxfunctions.ResourceReader;
import org.gvsig.mobile.extension.wms.fmap.drivers.wms.FMapWMSDriver;
import org.gvsig.mobile.extension.wms.fmap.drivers.wms.FMapWMSDriverFactory;
import org.gvsig.mobile.extension.wms.fmap.drivers.wms.WMSException;
import org.gvsig.remoteclient.wms.WMSStatus;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.dynobject.DynStruct;
import org.gvsig.tools.persistence.PersistenceManager;
import org.gvsig.tools.persistence.PersistentState;
import org.gvsig.tools.persistence.exception.PersistenceException;
import org.gvsig.tools.task.Cancellable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * FMap's WMS Layer class.
 * 
 * Las capas WMS son tileadas para descargarlas del servidor. Esto quiere decir
 * que estn formadas por multiples ficheros raster. Por esto la fuente de datos
 * raster (IRasterDatasource) de la capa FLyrWMS es un objeto de tipo
 * CompositeDataset. Este objeto est compuesto por un array bidimensional de
 * MultiRasterDataset. Cada uno de los MultiRasterDataset corresponde con un
 * tile salvado en disco. Estos MultiRasterDataset se crean cada vez que se
 * repinta ya que en WMS a cada zoom varian los ficheros fuente. La secuencia de
 * creacin de un CompositeDataset sera la siguiente:
 * <UL>
 * <LI>Se hace una peticin de dibujado por parte del usuario llamando al mtodo
 * draw de FLyrWMS</LI>
 * <LI>Se tilea la peticin</LI>
 * <LI>Cada tile se dibuja abriendo una FLyerRaster para ese tile</LI>
 * <LI>Si es el primer dibujado se guarda una referencia en la capa WMS a las
 * propiedades de renderizado, orden de bandas, transparencia, filtros
 * aplicados, ...</LI>
 * <LI>Si no es el primer dibujado se asignan las propiedades de renderizado
 * cuya referencia se guarda en la capa WMS</LI>
 * <LI>Se guarda el MultiRasterDataset de cada tile</LI>
 * <LI>Al acabar todos los tiles creamos un CompositeDataset con los
 * MultiRasterDataset de todos los tiles</LI>
 * <LI>Asignamos a la capa la referencia de las propiedades de renderizado que
 * tenemos almacenadas. De esta forma si hay alguna modificacin desde el cuadro
 * de propiedades ser efectiva sobre los tiles que se dibujan.</LI>
 * </UL>
 * 
 * 
 * @author Jaume Dominguez Faus
 * 
 */
public class FLyrWMS extends FLyrDefault implements IHasImageLegend {
	private static final GeometryManager geomManager = GeometryLocator
			.getGeometryManager();
	private static final Logger logger = LoggerFactory.getLogger(FLyrWMS.class);
	private boolean mustTileDraw = true;
	private final int maxTileDrawWidth = 1023;
	private final int maxTileDrawHeight = 1023;

	public URL host;
	public String m_Format;

	private String m_SRS;
	private String layerQuery;
	private String infoLayerQuery;
	private FMapWMSDriver wms;
	private WMSStatus wmsStatus = new WMSStatus();
	private boolean wmsTransparency;
	private Vector styles;
	private Vector dimensions;
	private Hashtable onlineResources = new Hashtable();
	private Dimension fixedSize;
	private VisualStatus visualStatus = new VisualStatus();
	private boolean queryable = false;

	private List disableUpdateDrawVersion;

	// FIXME DynValue???
	private Envelope fullEnvelope = null;

	public FLyrWMS() {
		super();
		this.updateDrawVersion();
	}

	public FLyrWMS(Map args) throws LoadLayerException {
		this.updateDrawVersion();

		String _host = (String) args.get("host");
		String _sLayer = (String) args.get("layer");
		Envelope full_enve = (Envelope) args.get("layerExtent");
		String _sSRS = (String) args.get("srs");
		String _sFormat = (String) args.get("format");
		// int _height = Integer.parseInt(args.get("height").toString());
		// int _width = Integer.parseInt(args.get("width").toString());

		init(_host,_sLayer,full_enve,_sSRS,_sFormat);
	}

	private void init(String _host, String _layer, Envelope _dataEnvelope,
			String _ssrs, String _format) throws LoadLayerException {
		
		FMapWMSDriver drv = null;
		String[] sLayers = CompatLocator.getStringUtils().split(_layer, ",");

		try {
			this.setHost(new URL(_host));
		} catch (MalformedURLException e) {
			// e.printStackTrace();
			throw new URLLayerException(getName(), e);
		}
		try {
			drv = this.getDriver();
		} catch (IllegalStateException e) {
			throw new LoadLayerException(getName(), e);
		} catch (IOException e) {
			throw new ConnectionErrorLayerException(getName(), e);
		}

		if (_format == null || _ssrs == null || _dataEnvelope == null) {
			if (!drv.connect(null)) {
				throw new ConnectionErrorLayerException(getName(), null);
			}

			WMSLayerNode wmsNode = drv.getLayer(sLayers[0]);

			if (wmsNode == null) {
				throw new LoadLayerException(getName(), null);
			}
			
			if (_format == null) {
				_format = this.getGreatFormat(drv.getFormats());
			}

			// SRS
			Vector allSrs = wmsNode.getAllSrs();
			boolean isSRSSupported = false;
			if (_ssrs != null) {
				for (int i = 0; i < allSrs.size(); i++) {
					if (((String) allSrs.get(i)).compareTo(_ssrs) == 0) {
						isSRSSupported = true;
					}
				}
			}

			if (!isSRSSupported) {
				for (int i = 0; i < allSrs.size(); i++) {
					if (((String) wmsNode.getAllSrs().get(i))
							.compareTo("EPSG:4326") == 0) {
						_ssrs = (String) wmsNode.getAllSrs().get(i);
					}
				}
				if (_ssrs == null) {
					_ssrs = (String) wmsNode.getAllSrs().get(0);
				}
			}
			
			if (_dataEnvelope == null) {
				try {
					Rectangle2D ext = drv.getLayersExtent(sLayers, _ssrs);
					_dataEnvelope = GeometryLocator.getGeometryManager().createEnvelope(
							ext.getMinX(), ext.getMinY(),
							ext.getMaxX(), ext.getMaxY(),
							SUBTYPES.GEOM2D);
					// Dimension imgdim = new Dimension(_width, _height);
					// _dataEnvelope = MapControl.getExtentFromFullExtentAndImageSize(_dataEnvelope, imgdim, true);
				} catch (Exception e) {
					throw new LoadLayerException("[Layers: " + _layer + " @ " + _host + "]", e);
				}
			}
		}

		this.setFullEnvelope(_dataEnvelope);
		this.setFormat(_format);
		this.setLayerQuery(_layer);
		this.setInfoLayerQuery("");
		this.setSRS(_ssrs);
		this.setName(_layer);

		if (_format.toLowerCase().indexOf("png") != -1) {
			wmsStatus.setTransparency(true);
		}
		
		wmsStatus.addLayerName(_layer);
		wmsStatus.setSrs(_ssrs);
		// wmsStatus.setHeight(_height);
		// wmsStatus.setWidth(_width);
		wmsStatus.setFormat(_format);

		// this.setOnlineResources(drv.getOnlineResources());
		load();
	}

	/**
	 * It choose the best format to load different maps if the server supports
	 * it. This format could be png, because it supports transparency.
	 * 
	 * @param formats
	 *            Arraywith all the formats supported by the server
	 * @return
	 */
	private String getGreatFormat(Vector formats) {
		for (int i = 0; i < formats.size(); i++) {
			String format = (String) formats.get(i);
			if (format.equals("image/jpg")) {
				return format;
			}
			if (format.equals("image/jpeg")) {
				return format;
			}
		}

		return (String) formats.get(0);
	}

	/**
	 * Clase que contiene los datos de visualizacin de WMS.
	 * 
	 * @author Nacho Brodin (brodin_ign@gva.es)
	 */
	private class VisualStatus {
		/**
		 * Ancho y alto de la imagen o del conjunto de tiles si los tiene.
		 * Coincide con el ancho y alto del viewPort
		 */
		private double minX = 0D;
		private double minY = 0D;
		private double maxX = 0D;
		private double maxY = 0D;
		/**
		 * Lista de nombre de fichero que componen toda la visualizacin.
		 */
		private String[] fileNames = null;

		public Object clone() {
			VisualStatus s = new VisualStatus();
			s.maxX = maxX;
			s.maxY = maxY;
			s.minX = minX;
			s.minY = minY;
			s.fileNames = new String[fileNames.length];
			for (int i = 0; i < fileNames.length; i++)
				s.fileNames[i] = fileNames[i];
			return s;
		}
	}

	/*
	 * 
	 * @see
	 * com.iver.cit.gvsig.fmap.layers.FLayer#draw(java.awt.image.BufferedImage,
	 * java.awt.Graphics2D, com.iver.cit.gvsig.fmap.ViewPort,
	 * com.iver.cit.gvsig.fmap.operations.Cancellable)
	 */
	private int callCount; // mess code, represents the amount of times the
							// methods drawFixedSize or drawTile where tried for
							// an extent
	private static final int MAX_RETRY_TIMES = 5; // mess code, represents the
													// max amount of retries
													// allowed.

	public void draw(BufferedImage image, Graphics2D g, ViewPort viewPort,
			Cancellable cancel, double scale) throws ReadException {

		MyCancellable mc = new MyCancellable(cancel);
		try {
			wmsStatus.setExtent(viewPort.getExtent());
			wmsStatus.setWidth(viewPort.getImageWidth());
			wmsStatus.setHeight(viewPort.getImageHeight());
			File f = getDriver().getMap(wmsStatus, mc);
			
			Image img = ResourceReader.getImage(f);
			// image = ResourceReader.getResourceImage("test", "test.png");
			// .getImage(new File("\\test.jpg"));
			
			logger.debug("WMS - received image, imagee = " + img);
			if ((img != null) && (img.getWidth(null) != -1)) {
				logger.debug("WMS - received image, image width = " + img.getWidth(null));
				
				g.drawImage(img, 0, 0, null);
			} else {
				logger.debug("WMS - received image, image IS  NULL");				
			}
		} catch (UnsupportedVersionLayerException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IllegalStateException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (WMSException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		



	}

	/**
	 * Closes files and releases memory (pointers to null)
	 */
	private void closeAndFree() {

	}

	/**
	 * Carga y dibuja el raster usando la librera
	 * 
	 * @param filePath
	 *            Ruta al fichero en disco
	 * @param g
	 *            Graphics2D
	 * @param vp
	 *            ViewPort
	 * @param scale
	 *            Escala para el draw
	 * @param cancel
	 *            Cancelacin para el draw
	 * @throws ReadException
	 * @throws LoadLayerException
	 * @throws ReadException
	 */
	private void rasterProcess(String filePath, Graphics2D g, ViewPort vp,
			double scale, Cancellable cancel, int nLyr)
			throws LoadLayerException, ReadException {

	}

	/**
	 * Obtiene la extensin del fichero de georreferenciacin
	 * 
	 * @return String con la extensin del fichero de georreferenciacin
	 *         dependiendo del valor del formato obtenido del servidor. Por
	 *         defecto asignaremos un .wld
	 */
	private String getExtensionWorldFile() {
		String extWorldFile = ".wld";
		if (m_Format.equals("image/tif") || m_Format.equals("image/tiff")) {
			extWorldFile = ".tfw";
		}
		if (m_Format.equals("image/jpeg")) {
			extWorldFile = ".jpgw";
		}
		return extWorldFile;
	}

	/**
	 * Calcula el contenido del fichero de georreferenciacin de una imagen.
	 * 
	 * @param bBox
	 *            Tamao y posicin de la imagen (en coordenadas de usuario)
	 * @param sz
	 *            Tamao de la imagen en pixeles.
	 * @return el 'WorldFile', como String.
	 * @throws IOException
	 */
	public String getDataWorldFile(Rectangle2D bBox, Dimension sz)
			throws IOException {
		StringBuffer data = new StringBuffer();
		data.append((bBox.getMaxX() - bBox.getMinX()) / (sz.width - 1) + "\n");
		data.append("0.0\n");
		data.append("0.0\n");
		data.append("-" + (bBox.getMaxY() - bBox.getMinY()) / (sz.height - 1)
				+ "\n");
		data.append("" + bBox.getMinX() + "\n");
		data.append("" + bBox.getMaxY() + "\n");
		return data.toString();
	}

	public void _print(Graphics2D g, ViewPort viewPort, Cancellable cancel,
			double scale) throws ReadException {
		draw(null, g, viewPort, cancel, scale);
	}

	/**
	 * Devuelve el FMapWMSDriver.
	 * 
	 * @return FMapWMSDriver
	 * 
	 * @throws IllegalStateException
	 * @throws UnsupportedVersionLayerException
	 * @throws IOException
	 */
	public FMapWMSDriver getDriver() throws IllegalStateException,
			UnsupportedVersionLayerException, IOException {
		return FMapWMSDriverFactory.getFMapDriverForURL(host);
	}

	/**
	 * Devuelve el FMapWMSDriver.
	 * 
	 * @return FMapWMSDriver
	 * 
	 * @throws IllegalStateException
	 * @throws UnsupportedVersionLayerException
	 * @throws IOException
	 */
	public void setDriver(FMapWMSDriver drv) {
		// TODO: Comprobar cambio
		wms = drv;
		this.updateDrawVersion();
	}

	/**
	 * Devuelve el URL.
	 * 
	 * @return URL.
	 */
	public URL getHost() {
		return host;
	}

	/**
	 * Inserta el URL.
	 * 
	 * @param host
	 *            URL.
	 */
	public void setHost(URL host) {
		if (this.host == host) {
			return;
		}
		if (this.host != null && this.host.equals(host)) {
			return;
		}

		this.host = host;
		this.updateDrawVersion();
	}

	/**
	 * Devuelve la informacin de la consulta.
	 * 
	 * @return String.
	 */
	public String getInfoLayerQuery() {
		return infoLayerQuery;
	}

	/**
	 * Inserta la informaci�n de la consulta.
	 * 
	 * @param infoLayerQuery
	 *            String.
	 */
	public void setInfoLayerQuery(String infoLayerQuery) {
		this.infoLayerQuery = infoLayerQuery;
	}

	/**
	 * Devuelve la consulta.
	 * 
	 * @return String.
	 */
	public String getLayerQuery() {
		return layerQuery;
	}

	/**
	 * Inserta la consulta.
	 * 
	 * @param layerQuery
	 *            consulta.
	 */
	public void setLayerQuery(String layerQuery) {
		if (this.layerQuery == layerQuery) {
			return;
		}
		if (this.layerQuery != null && this.layerQuery.equals(layerQuery)) {
			return;
		}

		this.layerQuery = layerQuery;
		this.updateDrawVersion();
	}

	/**
	 * Devuelve el formato.
	 * 
	 * @return Formato.
	 */
	public String getFormat() {
		return m_Format;
	}

	/**
	 * Inserta el formato.
	 * 
	 * @param format
	 *            Formato.
	 */
	public void setFormat(String format) {
		if (this.m_Format == format) {
			return;
		}
		if (this.m_Format != null && this.m_Format.equals(format)) {
			return;
		}
		m_Format = format;
		this.updateDrawVersion();
	}

	/**
	 * Devuelve el SRS.
	 * 
	 * @return SRS.
	 */
	public String getSRS() {
		return m_SRS;
	}

	/**
	 * Inserta el SRS.
	 * 
	 * @param m_srs
	 *            SRS.
	 */
	public void setSRS(String m_srs) {
		if (m_SRS == m_srs) {
			return;
		}
		if (m_SRS != null && m_SRS.equals(m_srs)) {
			return;
		}
		m_SRS = m_srs;
		this.updateDrawVersion();
		setProjection(CRSFactory.getCRS(getSRS()));
	}

	/**
	 * Inserta la extensi�n total de la capa.
	 * 
	 * @param fullExtent
	 *            Rect�ngulo.
	 * @deprecated
	 */
	public void setFullExtent(Rectangle2D fullExtent) {
		this.setFullEnvelope(this.getEnvelopeFromRectable2D(fullExtent));
	}

	/**
	 * Inserta la extensi�n total de la capa en la proyeccion original.
	 * 
	 * @param fullExtent
	 *            Rect�ngulo.
	 */
	public void setFullEnvelope(Envelope envelope) {
		Envelope cur = null;
		try {
			cur = this.getFullEnvelope();
		} catch (ReadException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		if (cur == envelope) {
			return;
		}
		if (cur != null && cur.equals(envelope)) {
			return;
		}

		this.fullEnvelope = envelope;
		this.updateDrawVersion();
	}

	public HashMap getProperties() {
		HashMap info = new HashMap();

		String[] layerNames = CompatLocator.getStringUtils().split(
				getLayerQuery(), ",");
		Vector layers = new Vector(layerNames.length);
		try {
			if (getDriver().connect(null)) {
				for (int i = 0; i < layerNames.length; i++) {
					layers.add(i, getDriver().getLayer(layerNames[i]));
				}
				info.put("name", getName());
				info.put("selectedLayers", layers);
				info.put("host", getHost());
				info.put("srs", getSRS());
				info.put("format", getFormat());
				info.put("wmsTransparency", new Boolean(wmsTransparency));
				info.put("styles", styles);
				info.put("dimensions", dimensions);
				info.put("fixedSize", fixedSize);
				return info;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	public double getMaxX() {
		return visualStatus.maxX;
	}

	public double getMaxY() {
		return visualStatus.maxY;
	}

	public double getMinX() {
		return visualStatus.minX;
	}

	public double getMinY() {
		return visualStatus.minY;
	}

	/**
	 * @return Returns the wmsTransparency.
	 */
	public boolean isWmsTransparent() {
		return wmsTransparency;
	}

	/**
	 * @param wmsTransparency
	 *            The wmsTransparency to set.
	 */
	public void setWmsTransparency(boolean wmsTransparency) {
		if (this.wmsTransparency == wmsTransparency) {
			return;
		}
		this.wmsTransparency = wmsTransparency;
		this.updateDrawVersion();
	}

	/**
	 * @param styles
	 */
	public void setStyles(Vector styles) {
		if (this.styles == styles) {
			return;
		}
		if (this.styles != null && styles != null) {
			if (this.styles.containsAll(styles)
					&& this.styles.size() == styles.size()) {
				return;
			}
		}
		this.styles = styles;
		this.updateDrawVersion();
	}

	/**
	 * Sets the dimension vector that is a list of key-value pairs containing
	 * the name of the dimension and the value for it
	 * 
	 * @param dimensions
	 */
	public void setDimensions(Vector dimensions) {
		if (this.dimensions == dimensions) {
			return;
		}
		if (this.dimensions != null && dimensions != null) {
			if (this.dimensions.containsAll(dimensions)
					&& this.dimensions.size() == dimensions.size()) {
				return;
			}
		}
		this.dimensions = dimensions;
		this.updateDrawVersion();
	}

	/**
	 * Sets the set of URLs that should be accessed for each operation performed
	 * to the server.
	 * 
	 * @param onlineResources
	 */
	public void setOnlineResources(Hashtable onlineResources) {
		if (this.onlineResources == onlineResources) {
			return;
		}
		if (this.onlineResources != null
				&& this.onlineResources.equals(onlineResources)) {
			return;
		}

		this.onlineResources = onlineResources;
		this.updateDrawVersion();
	}

	/**
	 * Gets the URL that should be accessed for an operation performed to the
	 * server.
	 * 
	 * @param onlineResources
	 */
	public String getOnlineResource(String operation) {
		return ((String) onlineResources.get(operation));
	}

	/**
	 * When a server is not fully featured and it only can serve constant map
	 * sizes this value must be set. It expresses the size in pixels (width,
	 * height) that the map will be requested.
	 * 
	 * @param Dimension
	 *            sz
	 */
	public void setFixedSize(Dimension sz) {
		if (this.fixedSize == sz) {
			return;
		}
		if (this.fixedSize != null && this.fixedSize.equals(sz)) {
			return;
		}
		fixedSize = sz;
		this.updateDrawVersion();
	}

	/**
	 * Tells whether if this layer must deal with the server with the
	 * constant-size limitations or not.
	 * 
	 * @return boolean.
	 */
	private boolean isSizeFixed() {
		return fixedSize != null && fixedSize.width > 0 && fixedSize.height > 0;
	}

	/**
	 * If it is true, this layer accepts GetFeatureInfo operations. This WMS
	 * operations maps to FMap's infoByPoint(p) operation.
	 * 
	 * @param b
	 */
	public void setQueryable(boolean b) {
		queryable = b;
	}

	public String getTocImageIcon() {
		return "icon-layer-wms";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.iver.cit.gvsig.fmap.layers.RasterOperations#getTileSize()
	 */
	public int[] getTileSize() {
		int[] size = { maxTileDrawWidth, maxTileDrawHeight };
		return size;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.iver.cit.gvsig.fmap.layers.RasterOperations#isTiled()
	 */
	public boolean isTiled() {
		return mustTileDraw;
	}

	public Image getImageLegend() {
		return null;
		// try {
		// if (wms == null) {
		// wms = getDriver();
		// }
		// if (wms.hasLegendGraphic()) {
		// wmsStatus.setOnlineResource((String) onlineResources
		// .get("GetLegendGraphic"));
		// String path = getPathImage();// File legend =
		// // getDriver().getLegendGraphic(wmsStatus,
		// // layerQuery, null);
		// Image img = null;
		// if ((path != null) && (path.length() > 0)) {
		// img = new Image(path).getImage();
		// }
		// return img;
		// }
		// } catch (Exception e) {
		// }
		// return null;
	}

	public String getPathImage() {
		try {
			File legend = getDriver().getLegendGraphic(wmsStatus, layerQuery,
					null);
			return legend.getAbsolutePath();
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	// /* (non-Javadoc)
	// * @see com.iver.cit.gvsig.fmap.layers.FLayer#newComposedLayer()
	// */
	// public ComposedLayer newComposedLayer() {
	// Preferences prefs = Preferences.userRoot().node("gvsig.wms");
	//
	// /*
	// * from java.util.prefs import Preferences
	// * prefs = Preferences.userRoot().node("gvsig.wms")
	// * prefs.put("useComposedLayer","true")
	// */
	//
	// String str = prefs.get("useComposedLayer","false");
	// Boolean useComposedLayer = Boolean.TRUE; // por defecto ya se usan
	// try {
	// useComposedLayer = Boolean.valueOf(str);
	// } catch (Exception e) {
	//
	// }
	// if (useComposedLayer.booleanValue()) {
	// return new ComposedLayerWMS();
	// } else {
	// return null;
	// }
	// }

	/**
	 * @param styles
	 */
	public Vector getStyles() {
		return this.styles;
	}

	/*
	 * Checks if can make a single petition for the two layers to the server
	 * 
	 * @see
	 * com.iver.cit.gvsig.fmap.layers.ComposedLayerWMS#canAdd(com.iver.cit.gvsig
	 * .fmap.layers.FLayer)
	 */
	boolean isComposedLayerCompatible(FLayer layer) {
		FLyrWMS aLayer;

		if (!(layer instanceof FLyrWMS)) {
			return false;
		}
		aLayer = (FLyrWMS) layer;
		if (!this.getHost().equals(aLayer.getHost())) {
			return false;
		}
		if (!this.getFormat().equals(aLayer.getFormat())) {
			return false;
		}
		if (!this.getSRS().equals(aLayer.getSRS())) {
			return false;
		}
		if (this.getInfoLayerQuery() != null) {
			if (!this.getInfoLayerQuery().equals(aLayer.getInfoLayerQuery())) {
				return false;
			}
		} else if (aLayer.getInfoLayerQuery() != null) {
			return false;
		}

		// isFixedSize es privado
		if ((this.fixedSize != null) && (aLayer.fixedSize != null)) {
			if (this.fixedSize.equals(aLayer.fixedSize)) {
				return false;
			}
		} else if ((this.fixedSize != null) != (aLayer.fixedSize != null)) {
			return false;
		}

		// time elevation (dimensions)
		if ((this.dimensions != null) && (aLayer.dimensions != null)) {
			if (this.dimensions.size() != aLayer.dimensions.size()) {
				return false;
			} else {
				Iterator iter = this.dimensions.iterator();
				while (iter.hasNext()) {
					if (!aLayer.dimensions.contains(iter.next())) {
						return false;
					}
				}
			}

		} else if ((this.dimensions != null) != (aLayer.dimensions != null)) {
			return false;
		}

		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.gvsig.fmap.raster.layers.FLyrRasterSE#getLegend()
	 */
	public ILegend getLegend() {
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.gvsig.fmap.raster.layers.FLyrRasterSE#overviewsSupport()
	 */
	public boolean overviewsSupport() {
		return false;
	}

	protected void updateDrawVersion() {
		if (this.disableUpdateDrawVersion != null) {

			Thread curThread = Thread.currentThread();

			Thread aThread;

			Iterator iter = this.disableUpdateDrawVersion.iterator();
			while (iter.hasNext()) {
				aThread = (Thread) ((WeakReference) iter.next()).get();
				if (aThread == null) {
					iter.remove();
				} else if (aThread.equals(curThread)) {
					return;
				}
			}
		}
		// Exception ex = new Exception();
		// ex.printStackTrace();
		super.updateDrawVersion();
	}

	protected void disableUpdateDrawVersion() {
		if (this.disableUpdateDrawVersion == null) {
			this.disableUpdateDrawVersion = new ArrayList();
		}
		this.disableUpdateDrawVersion.add(new WeakReference(Thread
				.currentThread()));
	}

	protected void enableUpdateDrawVersion() {
		if (this.disableUpdateDrawVersion == null) {
			return;
		}

		Thread curThread = Thread.currentThread();

		Thread aThread;

		Iterator iter = this.disableUpdateDrawVersion.iterator();
		while (iter.hasNext()) {
			aThread = (Thread) ((WeakReference) iter.next()).get();
			if (aThread == null) {
				iter.remove();
			} else if (aThread.equals(curThread)) {
				iter.remove();
				break;
			}
		}
	}

	private Rectangle2D.Double getRectable2DFromEnvelope(Envelope env) {
		return new Rectangle2D.Double(env.getMinimum(0), env.getMinimum(1), env
				.getLength(0), env.getLength(1));
	}

	private Envelope getEnvelopeFromRectable2D(Rectangle2D rect) {
		try {
			return geomManager.createEnvelope(rect.getMinX(), rect.getMinY(),
					rect.getMaxX(), rect.getMaxY(), SUBTYPES.GEOM2D);
		} catch (CreateEnvelopeException e) {
			logger.error("Error creating the envelope", e);
		}
		return null;
	}

	/**
	 * Devuelve el envelope en la proyeccion de la vista
	 * 
	 */
	public Envelope getFullEnvelope() throws ReadException {
		Envelope rAux;
		rAux = this.fullEnvelope;

		// Esto es para cuando se crea una capa nueva con el fullExtent de ancho
		// y alto 0.
		if (rAux == null || rAux.getMaximum(0) - rAux.getMinimum(0) == 0
				&& rAux.getMaximum(1) - rAux.getMinimum(1) == 0) {
			try {
				rAux = geomManager.createEnvelope(0, 0, 100, 100,
						SUBTYPES.GEOM2D);
			} catch (CreateEnvelopeException e) {
				logger.error("Error creating the envelope", e);
			}
		}
		// Si existe reproyeccin, reproyectar el extent
//		ICoordTrans ct = getCoordTrans();
//		if (ct != null) {
//			Point2D pt1 = new Point2D.Double(rAux.getMinimum(0), rAux
//					.getMinimum(1));
//			Point2D pt2 = new Point2D.Double(rAux.getMaximum(0), rAux
//					.getMaximum(1));
//			pt1 = ct.convert(pt1, null);
//			pt2 = ct.convert(pt2, null);
//			try {
//				rAux = geomManager.createEnvelope(pt1.getX(), pt1.getY(), pt2
//						.getX(), pt2.getY(), SUBTYPES.GEOM2D);
//			} catch (CreateEnvelopeException e) {
//				logger.error("Error creating the envelope", e);
//			}// new
//			// Rectangle2D.Double();
//		}
		return rAux;

	}

	public void print(Graphics2D g, ViewPort viewPort, Cancellable cancel,
			double scale, PrintAttributes properties) throws ReadException {
		// TODO Auto-generated method stub

	}

	public DynStruct[] getMetadataChildren(){
		// TODO Auto-generated method stub
		return null;
	}

	public Object getMetadataID() {
		// TODO Auto-generated method stub
		return null;
	}

	public String getMetadataName() {
		// TODO Auto-generated method stub
		return null;
	}

	private class MyCancellable implements org.gvsig.compat.net.ICancellable {
		private Cancellable original;

		public MyCancellable(Cancellable cancelOriginal) {
			this.original = cancelOriginal;
		}

		public boolean isCanceled() {
			if (original == null) {
				return false;
			}
			return original.isCanceled();
		}

		public Object getID() {
			return this;
		}
	}
	
	public void doDispose() {
	}
	
	// =================================================
	
	public void loadFromState(PersistentState state) throws PersistenceException {
		
		super.loadFromState(state);
		
		String _host = state.getString("host");
		String _sLayer = state.getString("layer");
		Rectangle2D layerExt = (Rectangle2D) state.get("layerExtent");
		String _sSRS = state.getString("srs");
		String _sFormat = state.getString("format");
		// int _height = state.getInt("height");
		// int _width = state.getInt("width");

		Envelope enve = null;
		try {
				enve = GeometryLocator.getGeometryManager().createEnvelope(
						layerExt.getMinX(), layerExt.getMinY(),
						layerExt.getMaxX(), layerExt.getMaxY(),
						SUBTYPES.GEOM2D);
		} catch (Exception e) {
			throw new PersistenceException("[Layers: " + _sLayer + " @ " + _host + "]", e);
		}
		
		try {
			init(_host,_sLayer,enve,_sSRS,_sFormat);
		} catch (LoadLayerException e) {
			throw new PersistenceException("[Layers: " + _sLayer + " @ " + _host + "]", e);
		}
	}

	
	
	public void saveToState(PersistentState state) throws PersistenceException {
		
		super.saveToState(state);
		state.set("host", host.toString());
		state.set("layer", layerQuery);
		
		Envelope enve = null;
		try {
			enve = getFullEnvelope();
			Rectangle2D r = new Rectangle2D.Double(
					enve.getMinimum(0),
					enve.getMinimum(1),
					enve.getLength(0),
					enve.getLength(1));
			state.set("layerExtent", r);
			state.set("srs", getSRS());
			state.set("format", getFormat());
		} catch (ReadException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		
		// state.set("height", wmsStatus.getHeight());
		// state.set("width", wmsStatus.getWidth());
	}
	
	
	
	
	public static void registerPersistent() {
		DynStruct definition =
			 ToolsLocator.getPersistenceManager().addDefinition(FLyrWMS.class,
					"FLyrWMS",
					"FLyrWMS Persistent definition",
	                null,
	                null);
	      		 	
		 definition.extend(PersistenceManager.PERSISTENCE_NAMESPACE,"FLyrVect");
		 
		 //New attribute
		 definition.addDynFieldString("host").setMandatory(true);
		 definition.addDynFieldString("layer").setMandatory(true);
		 definition.addDynFieldObject("layerExtent").setMandatory(true);
		 definition.addDynFieldString("srs").setMandatory(true);
		 definition.addDynFieldString("format").setMandatory(true);     	
	}

}
