/**
 *     Caption: J2SE/J2ME(Personal Profile) Light-Weight Visual Components Library
 *
 *     @author:     Andrei Vishnevsky
 *     @version:    4.50
 *
 * The latest product version is always available from the product's homepage:
 * http://www.lwvcl.com/
 *
 * Copyright (C) 2001-2005  Zaval Creative Engineering Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.gvsig.mobile.ui.lw;

import java.awt.Color;
import java.awt.Insets;
import java.util.Properties;
import java.util.StringTokenizer;

import org.gvsig.i18n.Messages;
import org.zaval.data.event.TextEvent;
import org.zaval.data.event.TextListener;
import org.zaval.lw.LwAList;
import org.zaval.lw.LwBorderLayout;
import org.zaval.lw.LwCanvas;
import org.zaval.lw.LwCompList;
import org.zaval.lw.LwComponent;
import org.zaval.lw.LwConstraints;
import org.zaval.lw.LwFlowLayout;
import org.zaval.lw.LwGridLayout;
import org.zaval.lw.LwLayout;
import org.zaval.lw.LwPanel;
import org.zaval.lw.LwSlider;
import org.zaval.lw.LwToolkit;
import org.zaval.lw.LwView;
import org.zaval.lw.event.LwActionListener;
import org.zaval.lw.event.LwActionSupport;
import org.zaval.lw.mask.BasicMaskValidator;
import org.zaval.lw.mask.LwMaskTextField;
import org.zaval.lw.mask.MaskValidator;

/**
 *  This panel provides ability to customize color using different ways :
 *  <ul>
 *    <li>By the specifying RGB color components in the text fields.</li>
 *    <li>By the specifying RGB color components using sliders.</li>
 *    <li>
 *      By the selecting one of the predefined color. The predefined colors set is specified by
 *      the properties. The properties has the following format:
 *      <pre>
 *         colors.rows=[number of rows]
 *         colors.cols=[number of columns]
 *
 *         colors.row.[row number]=[red1],[green1],[blue1];[red2],[green2],[blue2]...[redN],[greenN],[blueN]
 *         ...
 *      </pre>
 *    </li>
 *  </ul>
 *  Register an action listener to be notified whenever a selected color has been changed.
 *  <p>
 *  The panel reads fields titles as the following LwVCL resources:
 *  <Table border="0">
 *    <tr>
 *      <td><b>cd.predef</b></td> <td>predefined color panel title</td>
 *    </tr>
 *    <tr>
 *      <td><b>cd.selected</b></td> <td>selected color field title</td>
 *    </tr>
 *    <tr>
 *      <td><b>cd.red</b></td> <td>red color field title</td>
 *    </tr>
 *    <tr>
 *      <td><b>cd.green</b></td> <td>green color field title</td>
 *    </tr>
 *    <tr>
 *      <td><b>cd.blue</b></td> <td>blue color field title</td>
 *    </tr>
 *  </table>
 *  </p>
 */
public class ColorPanel
extends LwPanel
implements LwActionListener, TextListener
{
  private LwActionSupport support;
  private LwMaskTextField redField, blueField, greenField;
  private LwSlider        redSlider, blueSlider, greenSlider;
  private LwCanvas        selected;
  private LwColorsList    colors;
  private Object          lock__;
  private LwView          bundle = LwToolkit.getView("cd.slb");

 /**
  * Constructs a color panel. In this case the predefined colors set is loaded by
  * "colors.properties" properties file.
  */
  public ColorPanel() {
    this(Color.red, "colors.properties");
  }

 /**
  * Constructs a color panel with the given selected color and the given properties file to
  * specify a predefined colors set. The properties file will be fetched relatively LwVCL base
  * (the current base can be got by the LwToolkit.getResourcesBase method) directory.
  * @param c the given selected color.
  * @param props the given properties file.
  */
  public ColorPanel(Color c, String props) {
    this(c, LwToolkit.getProperties(props));
  }

 /**
  * Constructs a color panel with the given selected color and the given properties to
  * specify a predefined colors set.
  * @param c the given selected color.
  * @param p the given properties.
  */
  public ColorPanel(Color c, Properties p)
  {	  
	this.setBackground(Color.white);
	
    setInsets(2,2,2,2);
    colors = new LwColorsList(p);
    colors.setColor (LwAList.SEL_COLOR, null);
    int boxSize = colors.getBoxSize();
    colors.addSelectionListener(this);
    
    LwPanel tPanel = new LwPanel();
    tPanel.setLwLayout(new LwBorderLayout());
    tPanel.add(LwBorderLayout.CENTER, colors);

    LwPanel sPanel = new LwPanel();    
    sPanel.setLwLayout(new LwGridLayout(1, 2, LwToolkit.HORIZONTAL));
    selected = new LwCanvas();
    selected.getViewMan(true).setBorder("br.plain");
    selected.setPSSize(2*boxSize, boxSize);
    sPanel.add (DialogConstants.getDefaultLabel("                "));
    sPanel.add (selected);

    LwPanel rgb = new LwPanel();
    MaskValidator rgbValidator = new BasicMaskValidator(0, 255);
    rgb.setLwLayout(new LwGridLayout(3, 3, LwToolkit.HORIZONTAL | LwToolkit.VERTICAL));
    LwConstraints rgbCon1 = new LwConstraints();
    rgbCon1.insets = new Insets(2, 2, 2, 2);
    rgbCon1.fill = LwToolkit.HORIZONTAL;
    LwConstraints rgbCon2 = new LwConstraints();
    rgbCon2.insets = new Insets(2, 2, 2, 2);
    rgbCon2.fill = LwToolkit.HORIZONTAL | LwToolkit.VERTICAL;

    rgb.add (rgbCon1, DialogConstants.getDefaultLabel(Messages.getText("ColorChooser.1")));
    redField = new LwMaskTextField("", "nnn");
    redField.setValidator(rgbValidator);
    redField.getTextModel().addTextListener(this);
    redField.getTextRender().setFont(DialogConstants.getDefaultFont());
    rgb.add (rgbCon1, redField);
    redSlider = new LwSlider();
    redSlider.showTitle(false);
    redSlider.showScale(false);
    redSlider.jumpOnPress(true);
//    redSlider.setView(LwSlider.BUNDLE_VIEW, bundle);
    redSlider.setValues(0, 255, new int[] { 0 }, 1, 1);
    redSlider.setPSSize(150, -1);
    redSlider.addActionListener(this);
    rgb.add (rgbCon2, redSlider);

    rgb.add (rgbCon1, DialogConstants.getDefaultLabel(Messages.getText("ColorChooser.2")));
    greenField = new LwMaskTextField("", "nnn");
    greenField.setValidator(rgbValidator);
    greenField.getTextModel().addTextListener(this);
    greenField.getTextRender().setFont(DialogConstants.getDefaultFont());
    rgb.add (rgbCon1, greenField);
    greenSlider = new LwSlider();
    greenSlider.showTitle(false);
    greenSlider.showScale(false);
    greenSlider.jumpOnPress(true);
//    greenSlider.setView(LwSlider.BUNDLE_VIEW, bundle);
    greenSlider.setValues(0, 255, new int[] { 0 }, 1, 1);
    greenSlider.setPSSize(150, -1);
    greenSlider.addActionListener(this);
    rgb.add (rgbCon2, greenSlider);

    rgb.add (rgbCon1, DialogConstants.getDefaultLabel(Messages.getText("ColorChooser.3")));
    blueField = new LwMaskTextField("", "nnn");
    blueField.setValidator(rgbValidator);
    blueField.getTextModel().addTextListener(this);
    blueField.getTextRender().setFont(DialogConstants.getDefaultFont());
    rgb.add (rgbCon1, blueField);
    blueSlider = new LwSlider();
    blueSlider.showTitle(false);
    blueSlider.showScale(false);
    blueSlider.jumpOnPress(true);
//    blueSlider.setView(LwSlider.BUNDLE_VIEW, bundle);
    blueSlider.setValues(0, 255, new int[] { 0 }, 1, 1);
    blueSlider.setPSSize(150, -1);
    blueSlider.addActionListener(this);
    rgb.add (rgbCon2, blueSlider);

    int w = Math.max(Math.max(sPanel.getPreferredSize().width, tPanel.getPreferredSize().width),
                              rgb.getPreferredSize().width);
    sPanel.setPSSize(w, -1);
    tPanel.setPSSize(w, -1);
    rgb.setPSSize(w, -1);
    
    add (sPanel);
    add (tPanel);
    add (rgb);
    
    sPanel.setBackground(Color.white);
//    tPanel.setBackground(Color.white);
    rgb.setBackground(Color.white);
    setSelectedColor(c);
  }

 /**
  * Adds the specified action listener to receive action events from this component.
  * The action event is performed whenever a selected color has been changed. The action event
  * stores the new color value in the event data field.
  * @param l the specified action listener.
  */
  public void addActionListener(LwActionListener l) {
    if (support == null) support = new LwActionSupport();
    support.addListener(l);
  }

 /**
  * Removes the specified action listener so it no longer receives action events
  * from this component.
  * @param l the specified action listener.
  */
  public void removeActionListener(LwActionListener l) {
    if (support != null) support.removeListener(l);
  }

  public void actionPerformed(Object target, Object data)
  {
    if (lock__ == null)
    {
      if (target == redSlider || target == greenSlider || target == blueSlider)
      {
        int r = redSlider.getValue(), g = greenSlider.getValue(), b = blueSlider.getValue();
        setRgbFields(r, g, b);
        colors.select(-1);
        changeBg(new Color(r, g, b));
      }
      else
      if (target == colors)
      {
        LwComponent c = (LwComponent)colors.getSelected();
        if (c != null)
        {
          Color color = c.getBackground();
          if (!color.equals(getSelectedColor()))
            syncByColor(color.getRed(), color.getGreen(), color.getBlue());
        }
      }
    }
  }

 /**
  * Sets the specified color as the selected color.
  * @param c the specified color to set as selected.
  */
  public void setSelectedColor(Color c) {
    Color oldColor = getSelectedColor();
    if (!oldColor.equals (c)) syncByColor (c.getRed(), c.getGreen(), c.getBlue());
  }

 /**
  * Gets the current selected color.
  * @return a current selected color.
  */
  public Color getSelectedColor() {
    return selected.getBackground();
  }

  public void textRemoved (TextEvent e) {
    fieldChanged();
  }

  public void textInserted (TextEvent e) {
    fieldChanged();
  }

  public void textUpdated (TextEvent e)  {
    fieldChanged();
  }

  protected /*C#override*/ LwLayout getDefaultLayout() {
    return new LwFlowLayout(LwToolkit.CENTER, LwToolkit.CENTER, LwToolkit.VERTICAL, 4);
  }

  private void fieldChanged()
  {
    if (lock__ == null)
    {
      int r = Integer.parseInt(redField.getText());
      int g = Integer.parseInt(greenField.getText());
      int b = Integer.parseInt(blueField.getText());
      setRgbSliders(r, g, b);
      colors.select(-1);
      changeBg(new Color(r, g, b));
    }
  }

  private void syncByColor(int r, int g, int b)
  {
    if (lock__ == null) lock__ = selected;
    try {
      setRgbFields(r, g, b);
      setRgbSliders(r, g, b);
      changeBg(new Color(r, g, b));
    }
    finally {
      if (lock__ == selected) lock__ = null;
    }
  }

  private void changeBg(Color c)  {
    selected.setBackground(c);
    if (support != null) support.perform (this, c);
  }

  private String completeInt(int v, int f)
  {
    StringBuffer buffer = new StringBuffer(Integer.toString(v));
    while (buffer.length() < f) buffer.insert(0, '0');
    return buffer.toString();
  }

  private void setRgbFields(int r, int g, int b)
  {
    if (lock__ == null) lock__ = redField;
    try {
      redField.setText(completeInt(r, 3));
      greenField.setText(completeInt(g, 3));
      blueField.setText(completeInt(b, 3));
    }
    finally {
      if (lock__ == redField) lock__ = null;
    }
  }

  private void setRgbSliders(int r, int g, int b)
  {
    if (lock__ == null) lock__ = redSlider;
    try {
      redSlider.setValue(r);
      greenSlider.setValue(g);
      blueSlider.setValue(b);
    }
    finally {
      if (lock__ == redSlider) lock__ = null;
    }
  }
}

class LwColorsList
extends LwCompList
{
  private int boxSize = 20;

  public LwColorsList(Properties p)
  {
    int rows = Integer.parseInt(p.getProperty("colors.rows"));
    int cols = Integer.parseInt(p.getProperty("colors.cols"));

    setLwLayout(new LwGridLayout(rows, cols, LwToolkit.HORIZONTAL | LwToolkit.VERTICAL));
    LwConstraints c1 = new LwConstraints();
    c1.insets = new Insets(3,3,3,3);
    this.setBackground(Color.white);

    for (int i=0; i<rows; i++)
    {
      /*C#char[]   seps   = {';', ','};*/
      /*C#string[] values = p.getProperty("colors.row." + i).Split(seps);*/
      /*C#for (int j=0;j<values.Length;j+=3)*/

      StringTokenizer st = new StringTokenizer(p.getProperty("colors.row." + i), ",;"); /*java*/
      while (st.hasMoreTokens())                                                        /*java*/
      {
        LwCanvas box = new LwCanvas();
        box.setBackground(new Color(Integer.parseInt(st.nextToken()),   /*java*/
                                    Integer.parseInt(st.nextToken()),   /*java*/
                                    Integer.parseInt(st.nextToken()))); /*java*/
        /*C#box.setBackground(new Color(System.Convert.ToInt32(values[j]),*/
        /*C#                            System.Convert.ToInt32(values[j+1]),*/
        /*C#                            System.Convert.ToInt32(values[j+2])));*/
        box.setPSSize(boxSize, boxSize);
        box.getViewMan(true).setBorder("br.sunken");
        add(c1, box);
      }
    }

    getViewMan(true).setBorder((LwView)null);
  }

  public int getBoxSize() {
    return boxSize;
  }

  //protected /*C#override*/ void drawSelMarker(Graphics g, int x, int y, int w, int h) {}

  protected /*C#override*/ int getItemGaps () {
    return 2;
  }
}













